# coding=utf-8
import datetime
import os
from collections import namedtuple

import torch
from mmf.common.registry import registry
from modeling_uniter import VisualBertConfig, VisualBertForQuestionAnswering
from sklearn.metrics import auc, roc_auc_score, roc_curve
from tqdm import tqdm
from transformers import AdamW  # LxmertForQuestionAnswering,
from transformers import get_linear_schedule_with_warmup

from dataloader import MMFLoader

torch.manual_seed(1)


def train(
    model, optim, train_data, valid_data, outfile, warmup=None, epochs=8, scheduler=None
):
    add_header = True
    for epoch in range(epochs):
        total = 0.0
        right = 0.0
        for b in tqdm(train_data):
            model.train()
            optim.zero_grad()
            b = train_data.toCuda(b)
            output = model(
                input_ids=b["input_ids"],
                visual_feats=b["roi_features"],
                visual_pos=b["boxes"],
                attention_mask=b["attention_mask"],
                labels=b["label"],
                token_type_ids=b["token_type_ids"],
                image_text_alignment=b["obj_input_ids"],
                extras=b.get("extras", None),
            )
            raise Exception(output.attentions)
            """

            d = cheat({
                'position_embeddings_visual': None,
                'visual_embeddings_type': None,
                'input_ids' : b["input_ids"],
                'visual_embeddings' : b["roi_features"],
                'visual_pos' : b["boxes"],
                'attention_mask' : b["attention_mask"],
                'segment_ids' : b["token_type_ids"],
                'input_mask': None,
                'image_text_alignment': None
            })
            output = model(d)
            """

            labels = b["label"]
            loss = output.loss
            # logit = output["scores"]
            logit = output.question_answering_score
            # loss = torch.nn.functional.cross_entropy(logit, labels)
            loss.backward()

            logit = torch.nn.functional.softmax(logit, dim=-1)
            logit, pred = logit.max(1)
            score = b["label"]
            right += (score.eq(pred.long())).sum()
            total += float(train_data.batch_size)
            # torch.nn..clip_grad_norm_(model.parameters(), 5.0)
            # norm = torch.nn..clip_grad_norm_(model.parameters(), 0.25)
            if warmup is not None:
                warmup.step()
            optim.step()
            for p in optim.param_groups:
                pass
                # print(p["lr"])

            if TEST:
                break

            if scheduler is not None:
                scheduler.step()

        val_acc = inference(model, valid_data)
        if add_header:
            add_header = False
            date = datetime.datetime.now()
            with open(outfile, "a") as f:
                f.write(f"\nName {NAME}\n")
                f.write(f"Date: {date} \n")
                f.flush()

        log_str = (
            f"\tEpoch {epoch}: train {(right/total*100):0.2f} % val"
            f" {(val_acc[0]):0.2f} % roc {(val_acc[1]):0.2f} %\n"
        )
        print(log_str)
        if not TEST:
            with open(outfile, "a") as f:
                f.write(log_str)
                f.flush()

    with open(outfile, "a") as f:
        date = datetime.datetime.now()
        if not TEST:
            f.write(f"Time: {date} \n")
        else:
            f.write(f"(TEST) Time: {date} \n")
        f.flush()

    total = 0.0
    right = 0.0
    preds = torch.Tensor([])
    scores = torch.Tensor([])
    for b in tqdm(loader):
        model.eval()
        b = loader.toCuda(b)
        output = model(
            input_ids=b["input_ids"],
            visual_feats=b["roi_features"],
            visual_pos=b["boxes"],
            attention_mask=b["attention_mask"],
            token_type_ids=b["token_type_ids"],
            image_text_alignment=b["obj_input_ids"],
        )
        """
        d = cheat({
            'position_embeddings_visual': None,
            'visual_embeddings_type': None,
            'input_ids' : b["input_ids"],
            'visual_embeddings' : b["roi_features"],
            'visual_pos' : b["boxes"],
            'attention_mask' : b["attention_mask"],
            'segment_ids' : b["token_type_ids"],
            'input_mask': None,
            'image_text_alignment': None
        })
        output = model(d)
        """

        logit = output.question_answering_score
        # logit = output["scores"]
        logit = torch.nn.functional.softmax(logit, dim=-1)
        score = b["label"]
        logit, pred = logit.max(-1)
        right += (score.eq(pred.long())).sum()
        total += float(loader.batch_size)
        scores = torch.cat((scores, score.detach().cpu()), dim=0)
        preds = torch.cat((preds, pred.detach().cpu()), dim=0)
    return right / total * 100, roc_auc_score(scores.numpy(), preds.numpy())


if __name__ == "__main__":

    outfile = os.path.realpath("p_logs.txt")
    # model = LxmertForQuestionAnswering.from_pretrained(
    #     os.path.realpath("language_heavy"
    # ))
    conf = VisualBertConfig()
    model = VisualBertForQuestionAnswering.from_pretrained(
        os.path.realpath("./hateful_ckp"),
    )

    # model = registry.get_model_class(
    #         "visual_bert"
    #     ).from_pretrained(
    #         "visual_bert.finetuned.hateful_memes"
    #     )
    # model.training_head_type = "finetuning"
    # model.resize_num_qa_labels(2)
    model = model.cuda()
    optim = AdamW(list(model.parameters()), 1e-04)
    # warmup = get_linear_schedule_with_warmup(optim, num_warmup_steps=1000,
    #         num_training_steps=22000)
    warmup = None
    scheduler = torch.optim.lr_scheduler.StepLR(
        optim, step_size=200, gamma=0.9, last_epoch=-1
    )
    # scheduler = None
    train_data = MMFLoader(
        "train.jsonl",
        "arrow/mmf.arrow",
        objs="objects.txt",
        max_objs=36,
        sent_length=20,
        batch_size=64,
        collate_pt=True,
        num_workers=8,
        percent=1,
    )
    val_data = MMFLoader(
        "dev.jsonl",
        "arrow/mmf.arrow",
        objs="objects.txt",
        max_objs=36,
        sent_length=20,
        batch_size=32,
        collate_pt=True,
        num_workers=0,
    )

    train(
        model,
        optim,
        train_data,
        val_data,
        outfile,
        warmup=warmup,
        epochs=6,
        scheduler=scheduler,
    )



import inspect
import os
import sys

import vltk.vars as vltk
from vltk.abc.adapter import Adapter
from vltk.abc.extraction import VisnExtraction
from vltk.abc.visnadapter import VisnDataset
from vltk.abc.visnlangadatper import VisnLangDataset
from vltk.inspection import get_classes


class Adapters:
    def __init__(self):
        if "ADAPTERDICT" not in globals():
            global ADAPTERDICT
            ADAPTERDICT = get_classes(vltk.ADAPTERS, Adapter, pkg="vltk.adapters")
        # top = inspect.stack()[-1][1]
        # name = "/".join(top.split("/")[:-1])
        # top = top.split("/")[-1].split(".")[0]
        # sys.path.append(name)
        # __import__(top)
        # clsmembers = inspect.getmembers(sys.modules[top])
        # raise Exception("woo", clsmembers)

    def is_visnlang(self, adapter: str):
        assert adapter in self.avail(), f"adapter {adapter} not is not available"
        adapter_class = self.get(adapter)
        return VisnLangDataset in adapter_class.__bases__

    def is_visn(self, adapter: str):
        assert adapter in self.avail(), f"adapter {adapter} not is not available"
        adapter_class = self.get(adapter)
        return VisnDataset in adapter_class.__bases__

    def is_extraction(self, adapter: str):
        assert adapter in self.avail(), f"adapter {adapter} not is not available"
        adapter_class = self.get(adapter)
        return VisnExtraction in adapter_class.__bases__

    @staticmethod
    def avail():
        return list(ADAPTERDICT.keys())

    def get(self, name):
        try:
            return ADAPTERDICT[name]
        except KeyError:
            raise Exception(f"{name} not available from {self.avail()}")

    def add(self, *args):
        for dset in args:
            ADAPTERDICT[dset.__name__.lower()] = dset



import torch
import vltk.vars as vltk
from PIL import Image
from vltk.features import Features
from vltk import adapters
from vltk.configs import VisionConfig
from vltk.utils.adapters import rescale_box


class FRCNN(adapters.VisnExtraction):

    # TODO: currently, this image preprocessing config is not correct
    default_processor = VisionConfig(
        **{
            "transforms": ["FromFile", "Resize", "ToTensor", "Normalize"],
            "size": 800,
            "max_size": 1333,
            "mode": Image.BILINEAR,
            "pad_value": 0.0,
            "mean": [102.9801 / 255, 115.9465 / 255, 122.7717 / 255],
            "std": [1.0, 1.0, 1.0],
        }
    )

    @staticmethod
    def setup():
        from vltk import compat
        from vltk.modeling.frcnn import FRCNN as FasterRCNN

        weights = "unc-nlp/frcnn-vg-finetuned"
        model_config = compat.Config.from_pretrained("unc-nlp/frcnn-vg-finetuned")
        return FasterRCNN.from_pretrained(weights, model_config), model_config

    @staticmethod
    def schema(max_detections=36, visual_dim=2048):
        return {
            "attr_ids": Features.Ids(),
            "object_ids": Features.Ids(),
            vltk.features: Features.Features3D(max_detections, visual_dim),
            vltk.box: Features.Box(),
        }

    @staticmethod
    def forward(model, entry):

        size = entry[vltk.size]
        scale_wh = entry[vltk.scale]
        image = entry[vltk.img]

        model_out = model(
            images=image.unsqueeze(0),
            image_shapes=size.unsqueeze(0),
            padding="max_detections",
            pad_value=0.0,
            location="cpu",
        )
        normalized_boxes = torch.round(rescale_box(model_out["boxes"][0], 1 / scale_wh))

        return {
            "object_ids": [model_out["obj_ids"][0].tolist()],
            "attr_ids": [model_out["attr_ids"][0].tolist()],
            vltk.box: [normalized_boxes.tolist()],
            vltk.features: [model_out["roi_features"][0]],
        }



import vltk.vars as vltk
from vltk.abc.processor import (LangProcessor, Processor, VisnLangProcessor,
                                VisnProcessor)
from vltk.inspection import get_classes


class Processors:
    def __init__(self):
        if "PROCESSORDICT" not in globals():
            global PROCESSORDICT
            PROCESSORDICT = get_classes(
                vltk.PROCESSORS, Processor, pkg="vltk.processing"
            )
            PROCESSORDICT.pop("visnlangprocessor", None)
            PROCESSORDICT.pop("langprocessor", None)
            PROCESSORDICT.pop("visnprocessor", None)

    def is_visnlang(self, processor: str):
        assert processor in self.avail(), f"adapter {processor} not is not available"
        processor_class = self.get(processor)
        return processor_class.__bases__[0] == VisnLangProcessor

    def is_visn(self, processor: str):
        assert processor in self.avail(), f"adapter {processor} not is not available"
        processor_class = self.get(processor)
        return processor_class.__bases__[0] == VisnProcessor

    def is_lang(self, processor: str):
        assert processor in self.avail(), f"adapter {processor} not is not available"
        processor_class = self.get(processor)
        return processor_class.__bases__[0] == LangProcessor

    @staticmethod
    def avail():
        return list(PROCESSORDICT.keys())

    def get(self, name):
        try:
            return PROCESSORDICT[name]
        except KeyError:
            raise Exception(f"{name} not available from {self.avail()}")

    def add(self, *args):
        for dset in args:
            PROCESSORDICT[dset.__name__.lower()] = dset



import json
import os
from collections import defaultdict

import vltk
from vltk.adapters import Adapters
from vltk.dataset.loader import VisionLanguageLoader, VisionLoader

_adapters = Adapters()


def download(config, datadir, name, _adapters, seen_download):
    path = os.path.join(datadir, name)
    if os.path.exists(path) and not config.redownload:
        return seen_download
    if name in seen_download:
        return seen_download
    if name not in seen_download:
        seen_download.add(name)
    _adapters.get(name).download(datadir)
    return seen_download


def extract(
    config, datadir, name, _adapters, extracted, extractor=None, annotations=False
):
    if extractor is not None:
        if extractor in extracted:
            return extracted[extractor]
    if name in extracted:
        return extracted[name]

    if extractor is not None:
        exists = os.path.exists(os.path.join(datadir, name, extractor))
        if not exists or config.reextract:
            is_data = _adapters.get(config.extractor).extract(
                datadir, dataset_name=name
            )
            extracted[config.extractor] = is_data
            return extracted
        else:
            return extracted
    else:
        if annotations:
            exists = os.path.exists(os.path.join(datadir, name, "annotations.arrow"))
            if not exists:
                exists = os.path.exists(
                    os.path.join(datadir, name, "annotations", "annotations.arrow")
                )
            if not exists or config.reextract:
                is_annotations = _adapters.get(name)
                # is_annotations = is_annotations.extract(datadir)
                # is_annotations = is_annotations.extract(datadir)
                try:
                    is_annotations = is_annotations.extract(datadir)
                except TypeError as e:
                    if "object is not iterable" not in e:
                        print(f"Warning: No Annotations for {name}.")
                    else:
                        raise Exception(e)
                extracted[name] = is_annotations
                return extracted
            else:
                return extracted
        else:
            exists = False
            for split in vltk.SPLITALIASES:
                temp = os.path.join(datadir, name, f"{split}.arrow")
                if os.path.exists(temp):
                    exists = True
            if not exists or config.reextract:
                visnlangdataset = _adapters.get(name).extract(datadir)
                extracted[name] = visnlangdataset
                return extracted
            else:
                return extracted


def init_datasets(config):
    train_loader = None
    eval_loader = None
    metadata_ids = None
    assert (
        config.lang.ignore_id < 0
    ), f"ignore id: {config.lang.ignore_id} must be negative"
    train_ds, eval_ds, to_load, datasets_type = parse_datasets(config)
    if datasets_type == vltk.VLDATA:
        out_dict = load_vl(to_load, train_ds, eval_ds, config)
        metadata_ids = out_dict["metadata_ids"]
        train = out_dict["train"]
        evalutation = out_dict["eval"]
        annos = out_dict["annotations"]
        visndatasetadapters = out_dict["visndatasetadapters"]

        if train:
            train_loader = VisionLanguageLoader(
                config,
                visnlangdatasetadapterdict=train,
                visndatasetadapterdict=visndatasetadapters,
                annotationdict=annos,
                metadata_ids=metadata_ids,
                is_train=True,
            )

        if evalutation:
            eval_loader = VisionLanguageLoader(
                config,
                visnlangdatasetadapterdict=evalutation,
                visndatasetadapterdict=visndatasetadapters,
                annotationdict=annos,
                metadata_ids=metadata_ids,
                is_train=False,
            )
    if datasets_type == vltk.VDATA:

        out_dict = load_v(to_load, train_ds, eval_ds, config)
        # if is_train is false
        train = out_dict["train"]
        evalutation = out_dict["eval"]
        metadata_ids = out_dict["metadata_ids"]
        annotations = out_dict["annotations"]

        if train:
            train_loader = VisionLoader(
                config,
                visndatasetadapterdict=train,
                annotationdict={k: v for k, v in annotations.items() if k in train_ds},
                metadata_ids=metadata_ids,
                is_train=True,
            )
        if evalutation:
            eval_loader = VisionLoader(
                config,
                visndatasetadapterdict=evalutation,
                annotationdict={k: v for k, v in annotations.items() if k in eval_ds},
                metadata_ids=metadata_ids,
                is_train=False,
            )

    # loaders = {
    # "train": train_loader if train_loader is not None else None,
    # "eval": eval_loader if eval_loader is not None else None,
    # }
    # loaders = [(k, v) for k, v in loaders.items()]
    # any_train = any(map(lambda x: x == "train", [k for k in loaders]))
    # loaders = sorted(loaders, key=lambda x: x[0], reverse=True)
    train_loader = train_loader if train_loader is not None else None
    eval_loader = eval_loader if eval_loader is not None else None
    if train_loader is not None and hasattr(train_loader.dataset, "tokenizer"):
        train_loader.tokenizer = train_loader.dataset.tokenizer
    if eval_loader is not None and hasattr(eval_loader.dataset, "tokenizer"):
        eval_loader.tokenizer = eval_loader.dataset.tokenizer

    return train_loader, eval_loader


def parse_datasets(config):
    load_visnlangdatasets = defaultdict(set)
    load_visndatasets = defaultdict(set)
    train_ds = defaultdict(set)
    eval_ds = defaultdict(set)
    train = config.train_datasets
    train = train if train is not None else []
    test = config.eval_datasets
    test = test if test is not None else []

    assert train or test, "Must specify dataset in config to instatiate"
    if train and isinstance(train[0], str):
        train = [train]
    if test and isinstance(test[0], str):
        test = [test]
    total = train + test
    all_img = False
    all_vl = False
    for pair in total:
        ds, split = pair[0], pair[1]
        ds = ds.lower()
        split = split.lower()
        splits = split_handler(split)
        # TODO: will need to change this
        if _adapters.is_visnlang(ds):
            all_vl = True
            load_visnlangdatasets[ds].update(splits)
        if _adapters.is_visn(ds):
            load_visndatasets[ds].update(splits)
            all_img = True
    for ds, split in train:
        train_ds[ds].update(split_handler(split))
    for ds, split in test:
        eval_ds[ds].update(split_handler(split))

    assert not (all_vl and all_img), "cannot specify mixture of VL and Vision datasets"
    datasets_type = vltk.VDATA if all_img else vltk.VLDATA
    to_load = load_visndatasets if all_img else load_visnlangdatasets
    return train_ds, eval_ds, to_load, datasets_type


def load_vl(to_load, train_ds, eval_ds, config):
    datadir = config.datadir
    loaded_eval = defaultdict(dict)  # will be datasetk
    loaded_train = defaultdict(dict)  # will be datasetk
    loaded_visndatasetadapters = defaultdict(dict)
    loaded_annotations = defaultdict(dict)
    metadata_ids = {}
    metadata_idxs = {}
    seen_download = set()
    extracted = {}
    for name in sorted(set(to_load.keys())):
        # -DOWNLOAD HERE
        seen_download = download(config, datadir, name, _adapters, seen_download)
        extracted = extract(config, datadir, name, _adapters, extracted)

        splits = split_handler(to_load[name])  # list looks like ['trainval', 'dev']
        for split in sorted(splits):
            # add visnlangdatasetadapter first
            if name not in extracted:
                visnlangdatasetadapter = _adapters.get(name).load(
                    datadir, split=split, config=config
                )
            else:
                visnlangdatasetadapter = extracted["name"]["split"]
            dataset_metadata = visnlangdatasetadapter.get_metadata_counters()
            for meta in dataset_metadata:
                if meta not in metadata_ids:
                    metadata_ids[meta] = {"": config.lang.ignore_id}
                    metadata_idxs[meta] = 0
                for l in sorted(dataset_metadata[meta].keys()):
                    if l not in metadata_ids[meta]:
                        metadata_ids[meta][l] = metadata_idxs[meta]
                        metadata_idxs[meta] += 1

            if name in eval_ds and split in split_handler(eval_ds[name]):
                loaded_eval[name][split] = visnlangdatasetadapter
            if name in train_ds and split in split_handler(train_ds[name]):
                loaded_train[name][split] = visnlangdatasetadapter
            print(f"Added VisnLangDataset {name}: {split}")
            # now add visndatasetadapter
            is_name, is_split = zip(*visnlangdatasetadapter.data_info[split].items())
            is_name = is_name[0]
            is_split = is_split[0][0]
            # first check to see if we want annotations
            if not config.ignore_annotations and is_name not in loaded_annotations:

                # -DOWNLOAD HERE
                seen_download = download(
                    config, datadir, is_name, _adapters, seen_download
                )
                extracted = extract(
                    config, datadir, name, _adapters, extracted, annotations=True
                )

                if name not in extracted:
                    try:
                        is_annotations = _adapters.get(is_name).load(
                            datadir, config=config
                        )
                    except Exception:
                        print(f"Warning: No Annotations for {is_name}")
                        is_annotations = _adapters.get(is_name)
                else:
                    is_annotations = extracted[name]
                loaded_annotations[is_name] = is_annotations

                try:
                    dataset_metadata = is_annotations.get_metadata_counters()
                    for meta in dataset_metadata:
                        if meta not in metadata_ids:
                            metadata_ids[meta] = {"": config.lang.ignore_id}
                            metadata_idxs[meta] = 0
                        for l in sorted(dataset_metadata[meta].keys()):
                            if l not in metadata_ids[meta]:
                                metadata_ids[meta][l] = metadata_idxs[meta]
                                metadata_idxs[meta] += 1
                except Exception:
                    pass

            if (
                is_name in loaded_visndatasetadapters[is_name]
                and is_split in loaded_visndatasetadapters[is_name]
            ):
                continue
            if config.extractor is not None:
                # this is if we want to get pre-computed features

                extracted = extract(
                    config, datadir, name, _adapters, extractor=config.extractor
                )

                if config.extractor not in extracted:
                    is_data = _adapters.get(config.extractor).load(
                        datadir, split=is_split, dataset_name=is_name, config=config
                    )
                else:
                    is_data = extracted[config.extractor]
            else:
                # this is if we want to get raw features (in the form {id: raw file})
                is_data = _adapters.get(is_name).load_imgid2path(
                    config.datadir, split=split, config=config
                )
                # print(is_data)
            if (
                is_name in loaded_visndatasetadapters
                and is_split in loaded_visndatasetadapters[is_name]
            ):
                pass
            else:
                loaded_visndatasetadapters[is_name][is_split] = is_data
                print(f"Added VisnDataset {is_name}: {is_split}")

    if config.metadata_filedict is not None:
        metadata_filedict = config.metadata_filedict
        for k in metadata_ids:
            if k in metadata_filedict:
                metadata_ids[k] = json.load(open(metadata_filedict[k]))

    return {
        "eval": loaded_eval,
        "train": loaded_train,
        "annotations": loaded_annotations,
        "visndatasetadapters": loaded_visndatasetadapters,
        "metadata_ids": metadata_ids,
    }


# provide overlap ids at some other point
def load_v(to_load, train_ds, eval_ds, config):
    loaded_eval = defaultdict(dict)  # will be dataset
    loaded_train = defaultdict(dict)  # will be dataset
    loaded_annotations = defaultdict(dict)
    metadata_ids = {}
    metadata_idxs = {}
    seen_download = set()
    extracted = {}
    for name in sorted(set(to_load.keys())):
        # -DOWNLOAD HERE
        seen_download = download(config, config.datadir, name, _adapters, seen_download)
        extracted = extract(
            config, config.datadir, name, _adapters, extracted, annotations=True
        )
        splits = split_handler(to_load[name])  # list looks like ['trainval', 'dev']
        if name in extracted:
            annotations = extracted[name]
        else:
            annotations = _adapters.get(name).load(config.datadir, config=config)
        loaded_annotations[name] = annotations
        for split in sorted(splits):
            imgids2pathes = _adapters.get(name).load_imgid2path(config.datadir, split)

            dataset_metadata = annotations.get_metadata_counters()
            for meta in dataset_metadata:
                if meta not in metadata_ids:
                    metadata_ids[meta] = {"": config.lang.ignore_id}
                    metadata_idxs[meta] = 0
                for l in sorted(dataset_metadata[meta].keys()):
                    if l not in metadata_ids[meta]:
                        metadata_ids[meta][l] = metadata_idxs[meta]
                        metadata_idxs[meta] += 1
            if name in eval_ds and split in split in eval_ds[name]:
                loaded_eval[name][split] = imgids2pathes
            if name in train_ds and split in train_ds[name]:
                loaded_train[name][split] = imgids2pathes
            print(f"Added VisnDatasetAdapter {name}: {split}")

    if config.metadata_filedict is not None:
        metadata_filedict = config.metadata_filedict
        for k in metadata_ids:
            if k in metadata_filedict:
                metadata_ids[k] = json.load(open(metadata_filedict[k]))

    return {
        "train": loaded_train,
        "eval": loaded_eval,
        "annotations": loaded_annotations,
        "metadata_ids": metadata_ids,
    }


def split_handler(splits):
    if isinstance(splits, str):
        splits = [splits]
    unique_splits = set()
    for split in splits:
        if split == "testdev":
            unique_splits.add(split)
        else:
            for valid in vltk.SPLITALIASES:
                if valid in split:
                    unique_splits.add(valid)
    if "" in splits:
        unique_splits.add("")
    return sorted(unique_splits)



import datasets as ds
from vltk.abc.imageset import Imageset


class Coco2014(Imageset):
    name = "coco2014"

    @staticmethod
    def default_features(max_detections, pos_dim, visual_dim):
        return {
            "attr_ids": ds.Sequence(length=max_detections, feature=ds.Value("float32")),
            "attr_probs": ds.Sequence(
                length=max_detections, feature=ds.Value("float32")
            ),
        }

    def forward(filepath, image_preprocessor, model, **kwargs):
        pass


coco = Coco2014.extract(
    name="coco2014",
    img_config=None,
    splits=[],
    searchdirs="/playpen1/home/avmendoz/data",
)

coco = Coco2014.from_file(
    "/playpen1/home/avmendoz/data/coco2014/boxes/annotations.arrow"
)

print(coco)



import vltk.vars as vltk
from vltk.features import Features
from vltk import compat
from vltk.adapters import VisnExtraction
from vltk.configs import ProcessorConfig
from vltk.modeling.frcnn import FRCNN as FasterRCNN


class FRCNN(VisnExtraction):

    default_processor = ProcessorConfig(
        **{
            "transforms": ["ToPILImage", "ToTensor", "ResizeTensor", "Normalize"],
            "size": (800, 1333),
            "mode": "bilinear",
            "pad_value": 0.0,
            "mean": [102.9801, 115.9465, 122.7717],
            "sdev": [1.0, 1.0, 1.0],
        }
    )
    # model_config = compat.Config.from_pretrained("unc-nlp/frcnn-vg-finetuned")
    weights = "unc-nlp/frcnn-vg-finetuned"
    model = FasterRCNN

    def schema(max_detections=36, visual_dim=2048):
        return {
            "attr_ids": Features.ids,
            "object_ids": Features.ids,
            vltk.features: Features.features(max_detections, visual_dim),
            vltk.box: Features.boxtensor(max_detections),
        }

    def forward(model, entry, **kwargs):

        size = entry["size"]
        scale_hw = entry["scale"]
        image = entry["image"]

        model_out = model(
            images=image.unsqueeze(0),
            image_shapes=size.unsqueeze(0),
            scales_yx=scale_hw.unsqueeze(0),
            padding="max_detections",
            pad_value=0.0,
            return_tensors="np",
            location="cpu",
        )
        return {
            "object_ids": model_out["obj_ids"],
            "attr_ids": model_out["attr_ids"],
            vltk.box: model_out["normalized_boxes"],
            vltk.features: model_out["roi_features"],
        }


if __name__ == "__main__":
    # dataset = FRCNN.extract("/home/eltoto/vltk/tests/visualgenome")
    dataset = FRCNN.load("/home/eltoto/vltk/tests/visualgenome", split="train")
    entry = dataset.get(dataset.imgids[1])
    for k, v in entry.items():
        print(k, type(v))
    print(dataset)



from subprocess import call

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


# anything we want to run in these classes we can, and setup.py will execute them
class Develop(develop):
    def run(self):
        try:
            call([""])
        except Exception:
            print("")
        super().run()


class Install(install):
    def run(self):
        try:
            call([""])
        except Exception:
            print("")
        super().run()


def dependencies() -> list:
    deps = []
    with open("./requirements.txt") as f:
        for l in f.readlines():
            if "-e" in l:
                continue
            deps.append(l)
    return deps


setup(
    name="vltk",
    version="1.0.3",
    # this command is to be used if we want to auto run any scripts
    # cmdclass={"develop": Develop, "install": Install},
    entry_points={"console_scripts": ["vltk = vltk.cli:main"]},
    author="Antonio Mendoza",
    author_email="antonio36764@gmail.com",
    description="The Vision-Language Toolkit (VLTK)",
    long_description=open("README.md").read(),
    packages=[
        "vltk",
        "tests",
        "vltk/utils",
        "vltk/modeling",
        "vltk/modeling/configs",
        "vltk/adapters/",
        "vltk/dataset",
        "vltk/processing",
        "vltk/abc",
    ],
    install_requires=dependencies(),
)



"""
 coding=utf-8
 Copyright 2018, Antonio Mendoza Hao Tan, Mohit Bansal, Huggingface team :)
 Adapted From Facebook Inc, Detectron2

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.import copy
 """

import copy
import fnmatch
import json
import os
import pickle as pkl
import shutil
import sys
import tarfile
import tempfile
from collections import OrderedDict
from contextlib import contextmanager
from functools import partial
from hashlib import sha256
from io import BytesIO
from pathlib import Path
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

import cv2
import numpy as np
import requests
import wget
from filelock import FileLock
from PIL import Image
from tqdm.auto import tqdm
from yaml import Loader, dump, load

try:
    import torch

    _torch_available = True
except ImportError:
    _torch_available = False


try:
    from torch.hub import _get_torch_home

    torch_cache_home = _get_torch_home()
except ImportError:
    torch_cache_home = os.path.expanduser(
        os.getenv(
            "TORCH_HOME", os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "torch")
        )
    )

default_cache_path = os.path.join(torch_cache_home, "transformers")

CLOUDFRONT_DISTRIB_PREFIX = "https://cdn.huggingface.co"
S3_BUCKET_PREFIX = "https://s3.amazonaws.com/models.huggingface.co/bert"
PATH = "/".join(str(Path(__file__).resolve()).split("/")[:-1])
CONFIG = os.path.join(PATH, "config.yaml")
ATTRIBUTES = os.path.join(PATH, "attributes.txt")
OBJECTS = os.path.join(PATH, "objects.txt")
PYTORCH_PRETRAINED_BERT_CACHE = os.getenv(
    "PYTORCH_PRETRAINED_BERT_CACHE", default_cache_path
)
PYTORCH_TRANSFORMERS_CACHE = os.getenv(
    "PYTORCH_TRANSFORMERS_CACHE", PYTORCH_PRETRAINED_BERT_CACHE
)
TRANSFORMERS_CACHE = os.getenv("TRANSFORMERS_CACHE", PYTORCH_TRANSFORMERS_CACHE)
WEIGHTS_NAME = "pytorch_model.bin"
CONFIG_NAME = "config.yaml"


def load_labels(objs=OBJECTS, attrs=ATTRIBUTES):
    vg_classes = []
    with open(objs) as f:
        for object in f.readlines():
            vg_classes.append(object.split(",")[0].lower().strip())

    vg_attrs = []
    with open(attrs) as f:
        for object in f.readlines():
            vg_attrs.append(object.split(",")[0].lower().strip())
    return vg_classes, vg_attrs


def load_checkpoint(ckp):
    r = OrderedDict()
    with open(ckp, "rb") as f:
        ckp = pkl.load(f)["model"]
    for k in copy.deepcopy(list(ckp.keys())):
        v = ckp.pop(k)
        if isinstance(v, np.ndarray):
            v = torch.tensor(v)
        else:
            assert isinstance(v, torch.tensor), type(v)
        r[k] = v
    return r


class Config:
    _pointer = {}

    def __init__(self, dictionary: dict, name: str = "root", level=0):
        self._name = name
        self._level = level
        d = {}
        for k, v in dictionary.items():
            if v is None:
                raise ValueError()
            k = copy.deepcopy(k)
            v = copy.deepcopy(v)
            if isinstance(v, dict):
                v = Config(v, name=k, level=level + 1)
            d[k] = v
            setattr(self, k, v)

        self._pointer = d

    def __repr__(self):
        return str(list((self._pointer.keys())))

    def __setattr__(self, key, val):
        self.__dict__[key] = val
        self.__dict__[key.upper()] = val
        levels = key.split(".")
        last_level = len(levels) - 1
        pointer = self._pointer
        if len(levels) > 1:
            for i, l in enumerate(levels):
                if hasattr(self, l) and isinstance(getattr(self, l), Config):
                    setattr(getattr(self, l), ".".join(levels[i:]), val)
                if l == last_level:
                    pointer[l] = val
                else:
                    pointer = pointer[l]

    def to_dict(self):
        return self._pointer

    def dump_yaml(self, data, file_name):
        with open(f"{file_name}", "w") as stream:
            dump(data, stream)

    def dump_json(self, data, file_name):
        with open(f"{file_name}", "w") as stream:
            json.dump(data, stream)

    @staticmethod
    def load_yaml(config):
        with open(config) as stream:
            data = load(stream, Loader=Loader)
        return data

    def __str__(self):
        t = "    "
        if self._name != "root":
            r = f"{t * (self._level-1)}{self._name}:\n"
        else:
            r = ""
        level = self._level
        for i, (k, v) in enumerate(self._pointer.items()):
            if isinstance(v, Config):
                r += f"{t * (self._level)}{v}\n"
                self._level += 1
            else:
                r += f"{t * (self._level)}{k}: {v} ({type(v).__name__})\n"
            self._level = level
        return r[:-1]

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        config_dict, kwargs = cls.get_config_dict(
            pretrained_model_name_or_path, **kwargs
        )
        return cls(config_dict)

    @classmethod
    def get_config_dict(cls, pretrained_model_name_or_path: str, **kwargs):

        cache_dir = kwargs.pop("cache_dir", None)
        force_download = kwargs.pop("force_download", False)
        resume_download = kwargs.pop("resume_download", False)
        proxies = kwargs.pop("proxies", None)
        local_files_only = kwargs.pop("local_files_only", False)

        if os.path.isdir(pretrained_model_name_or_path):
            config_file = os.path.join(pretrained_model_name_or_path, CONFIG_NAME)
        elif os.path.isfile(pretrained_model_name_or_path) or is_remote_url(
            pretrained_model_name_or_path
        ):
            config_file = pretrained_model_name_or_path
        else:
            config_file = hf_bucket_url(
                pretrained_model_name_or_path, filename=CONFIG_NAME, use_cdn=False
            )

        try:
            # Load from URL or cache if already cached
            resolved_config_file = cached_path(
                config_file,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=local_files_only,
            )
            # Load config dict
            if resolved_config_file is None:
                raise EnvironmentError

        except EnvironmentError:
            # try laoding from file
            try:
                config_file = Config.load_yaml(config_file)
            except Exception:
                msg = "Can't load config for"
                raise EnvironmentError(msg)

        if resolved_config_file == config_file:
            pass
        else:
            pass

        return Config.load_yaml(resolved_config_file), kwargs


# quick compare tensors
def compare(in_tensor):

    out_tensor = torch.load("dump.pt", map_location=in_tensor.device)
    n1 = in_tensor.numpy()
    n2 = out_tensor.numpy()[0]
    print(n1.shape, n1[0, 0, :5])
    print(n2.shape, n2[0, 0, :5])
    assert np.allclose(
        n1, n2, rtol=0.01, atol=0.1
    ), f"{sum([1 for x in np.isclose(n1, n2, rtol=0.01, atol=0.1).flatten() if x == False])/len(n1.flatten())*100:.4f} % element-wise mismatch"
    raise Exception("tensors are all good")

    # Hugging face functiions below


def is_remote_url(url_or_filename):
    parsed = urlparse(url_or_filename)
    return parsed.scheme in ("http", "https")


def hf_bucket_url(model_id: str, filename: str, use_cdn=True) -> str:
    endpoint = CLOUDFRONT_DISTRIB_PREFIX if use_cdn else S3_BUCKET_PREFIX
    legacy_format = "/" not in model_id
    if legacy_format:
        return f"{endpoint}/{model_id}-{filename}"
    else:
        return f"{endpoint}/{model_id}/{filename}"


def http_get(
    url,
    temp_file,
    proxies=None,
    resume_size=0,
    user_agent=None,
):
    ua = "python/{}".format(sys.version.split()[0])
    if _torch_available:
        ua += "; torch/{}".format(torch.__version__)
    if isinstance(user_agent, dict):
        ua += "; " + "; ".join("{}/{}".format(k, v) for k, v in user_agent.items())
    elif isinstance(user_agent, str):
        ua += "; " + user_agent
    headers = {"user-agent": ua}
    if resume_size > 0:
        headers["Range"] = "bytes=%d-" % (resume_size,)
    response = requests.get(url, stream=True, proxies=proxies, headers=headers)
    if response.status_code == 416:  # Range not satisfiable
        return
    content_length = response.headers.get("Content-Length")
    total = resume_size + int(content_length) if content_length is not None else None
    progress = tqdm(
        unit="B",
        unit_scale=True,
        total=total,
        initial=resume_size,
        desc="Downloading",
    )
    for chunk in response.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            progress.update(len(chunk))
            temp_file.write(chunk)
    progress.close()


def get_from_cache(
    url,
    cache_dir=None,
    force_download=False,
    proxies=None,
    etag_timeout=10,
    resume_download=False,
    user_agent=None,
    local_files_only=False,
):

    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    os.makedirs(cache_dir, exist_ok=True)

    etag = None
    if not local_files_only:
        try:
            response = requests.head(
                url, allow_redirects=True, proxies=proxies, timeout=etag_timeout
            )
            if response.status_code == 200:
                etag = response.headers.get("ETag")
        except (EnvironmentError, requests.exceptions.Timeout):
            # etag is already None
            pass

    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    # etag is None = we don't have a connection, or url doesn't exist, or is otherwise inaccessible.
    # try to get the last downloaded one
    if etag is None:
        if os.path.exists(cache_path):
            return cache_path
        else:
            matching_files = [
                file
                for file in fnmatch.filter(os.listdir(cache_dir), filename + ".*")
                if not file.endswith(".json") and not file.endswith(".lock")
            ]
            if len(matching_files) > 0:
                return os.path.join(cache_dir, matching_files[-1])
            else:
                # If files cannot be found and local_files_only=True,
                # the models might've been found if local_files_only=False
                # Notify the user about that
                if local_files_only:
                    raise ValueError(
                        "Cannot find the requested files in the cached path and outgoing traffic has been"
                        " disabled. To enable model look-ups and downloads online, set 'local_files_only'"
                        " to False."
                    )
                return None

    # From now on, etag is not None.
    if os.path.exists(cache_path) and not force_download:
        return cache_path

    # Prevent parallel downloads of the same file with a lock.
    lock_path = cache_path + ".lock"
    with FileLock(lock_path):

        # If the download just completed while the lock was activated.
        if os.path.exists(cache_path) and not force_download:
            # Even if returning early like here, the lock will be released.
            return cache_path

        if resume_download:
            incomplete_path = cache_path + ".incomplete"

            @contextmanager
            def _resumable_file_manager():
                with open(incomplete_path, "a+b") as f:
                    yield f

            temp_file_manager = _resumable_file_manager
            if os.path.exists(incomplete_path):
                resume_size = os.stat(incomplete_path).st_size
            else:
                resume_size = 0
        else:
            temp_file_manager = partial(
                tempfile.NamedTemporaryFile, dir=cache_dir, delete=False
            )
            resume_size = 0

        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with temp_file_manager() as temp_file:
            print(
                "%s not found in cache or force_download set to True, downloading to %s",
                url,
                temp_file.name,
            )

            http_get(
                url,
                temp_file,
                proxies=proxies,
                resume_size=resume_size,
                user_agent=user_agent,
            )

        os.replace(temp_file.name, cache_path)

        meta = {"url": url, "etag": etag}
        meta_path = cache_path + ".json"
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)

    return cache_path


def url_to_filename(url, etag=None):

    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    if url.endswith(".h5"):
        filename += ".h5"

    return filename


def cached_path(
    url_or_filename,
    cache_dir=None,
    force_download=False,
    proxies=None,
    resume_download=False,
    user_agent=None,
    extract_compressed_file=False,
    force_extract=False,
    local_files_only=False,
):
    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    if isinstance(url_or_filename, Path):
        url_or_filename = str(url_or_filename)
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    if is_remote_url(url_or_filename):
        # URL, so get it from the cache (downloading if necessary)
        output_path = get_from_cache(
            url_or_filename,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            user_agent=user_agent,
            local_files_only=local_files_only,
        )
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        output_path = url_or_filename
    elif urlparse(url_or_filename).scheme == "":
        # File, but it doesn't exist.
        raise EnvironmentError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError(
            "unable to parse {} as a URL or as a local path".format(url_or_filename)
        )

    if extract_compressed_file:
        if not is_zipfile(output_path) and not tarfile.is_tarfile(output_path):
            return output_path

        # Path where we extract compressed archives
        # We avoid '.' in dir name and add "-extracted" at the end: "./model.zip" => "./model-zip-extracted/"
        output_dir, output_file = os.path.split(output_path)
        output_extract_dir_name = output_file.replace(".", "-") + "-extracted"
        output_path_extracted = os.path.join(output_dir, output_extract_dir_name)

        if (
            os.path.isdir(output_path_extracted)
            and os.listdir(output_path_extracted)
            and not force_extract
        ):
            return output_path_extracted

        # Prevent parallel extractions
        lock_path = output_path + ".lock"
        with FileLock(lock_path):
            shutil.rmtree(output_path_extracted, ignore_errors=True)
            os.makedirs(output_path_extracted)
            if is_zipfile(output_path):
                with ZipFile(output_path, "r") as zip_file:
                    zip_file.extractall(output_path_extracted)
                    zip_file.close()
            elif tarfile.is_tarfile(output_path):
                tar_file = tarfile.open(output_path)
                tar_file.extractall(output_path_extracted)
                tar_file.close()
            else:
                raise EnvironmentError(
                    "Archive format of {} could not be identified".format(output_path)
                )

        return output_path_extracted

    return output_path


def get_data(query, delim=","):
    assert isinstance(query, str)
    if os.path.isfile(query):
        with open(query) as f:
            data = f.read()
            try:
                data = eval(data)
            except Exception:
                data = data.split("\n")
    else:
        req = requests.get(query)
        try:
            data = requests.json()
        except Exception:
            data = req.content.decode()
            assert data is not None, "could not connect"
            try:
                data = eval(data)
            except Exception:
                data = data.split("\n")
        req.close()
    return data


def get_image_from_url(url):
    response = requests.get(url)
    img = np.array(Image.open(BytesIO(response.content)))
    return img


# to load legace frcnn checkpoint from detectron
def load_frcnn_pkl_from_url(url):
    fn = url.split("/")[-1]
    if fn not in os.listdir(os.getcwd()):
        wget.download(url)
    with open(fn, "rb") as stream:
        weights = pkl.load(stream)
    model = weights.pop("model")
    new = {}
    for k, v in model.items():
        new[k] = torch.from_numpy(v)
        if "running_var" in k:
            zero = torch.Tensor([0])
            k2 = k.replace("running_var", "num_batches_tracked")
            new[k2] = zero
    return new


def get_demo_path():
    print(f"{os.path.abspath(os.path.join(PATH, os.pardir))}/demo.ipynb")


def img_tensorize(im, input_format="RGB"):
    assert isinstance(im, str)
    # if os.path.isfile(im):
    img = cv2.imread(im)
    if img is not None:
        img = img[:, :, ::1]
    return img


def chunk(images, batch=1):
    return (images[i : i + batch] for i in range(0, len(images), batch))



import inspect
import math
import os
import random
import resource
import sys

import torch
import vltk.vars as vltk
# disable logging from datasets
from datasets.utils.logging import set_verbosity_error
from vltk.dataset.basedataset import (CollatedVLSets, SplitRangesVision,
                                      SplitRangesVL)
from vltk.dataset.langdataset import LangDataset
from vltk.dataset.visndataset import VisionDataset
from vltk.processing import Processors, VisnLangProcessor

__import__("tokenizers")
TOKENIZERS = {
    m[0]: m[1] for m in inspect.getmembers(sys.modules["tokenizers"], inspect.isclass)
}

rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (6144, rlimit[1]))

set_verbosity_error()


os.environ["TOKENIZERS_PARALLELISM"] = "False"


# TODO
class VisionLanguageDataset(VisionDataset, LangDataset):
    visn = set()
    lang = set()

    def __init__(
        self,
        config,
        visnlangdatasetadapterdict,  # contains visnlang annotations
        visndatasetadapterdict,  # contains dict of files or dataset of features
        annotationdict=None,  # conatains annotations for vision datasets
        metadata_ids=None,
        is_train=False,
        batch_info=None,
        tokenizer_in_visn_dataset=False,
        replace_keys=None,
        **kwargs,
    ):
        # ======
        # checking/setting respective adatpers
        uniq_imgs, missing_ids, shrink_lang, shrink_vision = self._check_uniq_imgs(
            visndatasetadapterdict, visnlangdatasetadapterdict
        )
        visndatasetadapterdict, visnlangdatasetadapterdict = self._tighten_datasets(
            uniq_imgs,
            visndatasetadapterdict,
            visnlangdatasetadapterdict,
            missing_ids,
            shrink_lang,
            shrink_vision,
        )
        self.tokenizer_in_visn_dataset = tokenizer_in_visn_dataset

        self.visnlangdatasetadapterdict = visnlangdatasetadapterdict
        self.visndatasetadapterdict = visndatasetadapterdict
        self.vl_idx_organizer = SplitRangesVL(visnlangdatasetadapterdict)
        self.visn_idx_organizer = SplitRangesVision(visndatasetadapterdict)
        # raise Exception(
        #     self.visn_idx_organizer.uniq_imgs, self.vl_idx_organizer.uniq_imgs
        # )
        """
        WARNING:
            untested: ensure all image ids are unique across image ids of various
            datasets.

            the simplest thing to do in the future will be to add an "adjust image
            IDS" function in the vision adapter and not just the vision language adatper
            That way, a check can be performed to make sure both are lined up
        """
        self.uniq_imgs = self.visn_idx_organizer.imgs
        visnlangdatasetadapters = []
        for dset in self.visnlangdatasetadapterdict:
            for split in self.visnlangdatasetadapterdict[dset]:
                visnlangdatasetadapters.append(
                    self.visnlangdatasetadapterdict[dset][split]
                )
        self.datasets = CollatedVLSets(*visnlangdatasetadapters)
        self.annotationdict = annotationdict
        # ======

        # ======
        # set some other properties
        self.config = config
        self.replace_keys = replace_keys
        self.batch_info = batch_info
        self.is_train = is_train
        self.metadata_ids = metadata_ids
        splits = self._check_uniq_splits()
        self.splits = splits
        self.placeholders = {}
        self.max_spanning_cols = kwargs.get("max_spanning_cols")
        # ======

        # ======
        # do tokenizer stuff
        self._init_tokenizer(config.lang)
        self._init_annotation_dict(config, annotationdict)
        self._init_image_processor(config)
        self._init_vision_processors(config)
        self._init_lang_processors(config)
        self._init_visnlang_processors(config)

        # ======

    def update_visn_lang_keys(self, lang_entry, visn_entry):
        self.visn = self.visn.union(set(visn_entry.keys()))
        self.lang = self.lang.union(set(lang_entry.keys()))

    def _init_visnlang_processors(self, config):
        visnlang_processors = (
            config.visnlang_processors if config.visnlang_processors is not None else []
        )

        visnlang_processors = [
            x if not isinstance(x, str) else Processors().get(x)
            for x in visnlang_processors
        ]

        visnlang_processors = list(
            filter(lambda x: x.__bases__[0] == VisnLangProcessor, visnlang_processors)
        )

        self.visnlang_processors = [
            x(
                config=self.config,
                metadata_ids=self.metadata_ids,
                from_transformers=self.from_transformers,
                tokenizer=self.tokenizer,
            )
            for x in visnlang_processors
        ]

        self.visnlang_processor_keys = ()
        for x in self.visnlang_processor_keys:
            self.visnlang_processor_keys += tuple(x.keys)

    def run_visnlang_processors(self, lang_entry, visn_entry, img_first):
        for processor in self.visnlang_processors:
            lang_entry, visn_entry = processor(
                lang_entry, visn_entry, img_first=img_first
            )
        return lang_entry, visn_entry

    def _tighten_datasets(
        self,
        uniq_imgs,
        visndatasetadapterdict,
        visnlangdatasetadapterdict,
        missing_ids,
        shrink_lang,
        shrink_vision,
    ):
        if missing_ids is not None:
            print(f"resizing datasets to account for {missing_ids} missing image IDs")
            if shrink_lang:
                for dset in visnlangdatasetadapterdict:
                    for split in visnlangdatasetadapterdict[dset]:
                        visnlang = visnlangdatasetadapterdict[dset][split]
                        filtered_visnlang = visnlang.imgid_filter(uniq_imgs, True)
                        visnlangdatasetadapterdict[dset][split] = filtered_visnlang

            if shrink_vision:
                for is_name in visndatasetadapterdict:
                    for is_split in visndatasetadapterdict[is_name]:
                        try:
                            # for visionsets that are Dataset objects
                            imgset = visndatasetadapterdict[is_name][is_split]
                            filtered_imgset = imgset.imgid_filter(uniq_imgs, False)
                            # TODO: remove later once I actually end up testing with this
                            # assert filtered_imgset.check_imgid_alignment()
                            visndatasetadapterdict[is_name][is_split] = filtered_imgset
                        except Exception:
                            # for vision sets that are only a dictionary of imgids
                            imgsetdict = visndatasetadapterdict[is_name][is_split]
                            imgsetdict = dict(
                                filter(lambda x: x[0] in uniq_imgs, imgsetdict.items())
                            )

                            visndatasetadapterdict[is_name][is_split] = imgsetdict

        return visndatasetadapterdict, visnlangdatasetadapterdict

    def _check_uniq_imgs(self, visndatasetadapterdict, visnlangdatasetadapterdict):
        uniq_visn_imgs = set()
        for is_info in visndatasetadapterdict.items():
            is_name, is_split_dict = is_info
            for k, imgset in is_split_dict.items():
                try:
                    img_ids = imgset.imgids
                    # Adapters().get(imgset.dataset).adjust_imgid()
                    # raise Exception(k, imgset, imgset.dataset)
                except Exception:
                    img_ids = imgset.keys()

                img_ids = set(img_ids)

                uniq_visn_imgs = uniq_visn_imgs.union(img_ids)
        uniq_lang_imgs = set()
        uniq_imgs = set()
        for ts_name, ts_splits in visnlangdatasetadapterdict.items():
            for split_name, ts in visnlangdatasetadapterdict[ts_name].items():
                temp_uniq = ts.uniq_imgs
                uniq_lang_imgs = uniq_lang_imgs.union(uniq_lang_imgs, temp_uniq)
                uniq_imgs = uniq_imgs.union(uniq_visn_imgs.intersection(temp_uniq))
        if not uniq_imgs:
            raise Exception(
                f"""
                ERROR: there are no common image IDs between either language or vision datasets,
                you may want to rename them or check to see if this should be the case or implement
                the `adjust_imgid` function in the VisnLangAdapter. \n
                Vision Dataset Image ID example: {next(iter(uniq_visn_imgs))}
                Language Dataset Image ID example: {next(iter(uniq_lang_imgs))}
                """
            )
        missing_ids = None
        shrink_lang = False
        shrink_vision = False
        all_imgs = uniq_lang_imgs.union(uniq_visn_imgs)
        if len(uniq_imgs) < len(all_imgs):
            missing_ids = len(all_imgs) - len(uniq_imgs)
            if not len(uniq_imgs) == len(uniq_visn_imgs):
                shrink_vision = True
            if not len(uniq_imgs) == len(uniq_lang_imgs):
                shrink_lang = True
        return uniq_imgs, missing_ids, shrink_lang, shrink_vision

    def _check_uniq_splits(self):
        splits = set()
        for v in self.visnlangdatasetadapterdict.values():
            splits = splits.union(set(v.keys()))
        return splits

    def _do_map_img_first(self, i):
        img_id = self.uniq_imgs[i]
        text_info = self.datasets.get(img_id)
        # try:
        # except Exception:
        #     for a in self.datasets.args:
        #         print(set(a["imgid"]))
        #     print("---")
        #     text_info = self.datasets.get(str(img_id))
        #     raise Exception(self.uniq_imgs, str(img_id))
        text_info.pop(vltk.imgid)
        text_info = self._handle_text_annotations(text_info, encode_batch=True)
        return text_info, img_id

    def _do_map_text_first(self, i):
        text_info = self.datasets[i]
        img_id = text_info[vltk.imgid]
        text_info = self._handle_text_annotations(text_info, encode_batch=False)
        return text_info, img_id

    # def random_visn_feat(self):
    #     rand_ind = random.randint(0, len(self.uniq_imgs) - 1)
    #     img_id = self.uniq_imgs[rand_ind]
    #     ts_name, ts_split = self.img2visnlangdatasetadapter[img_id]
    #     visnlangdatasetadapter = self.visnlangdatasetadapterdict[ts_name][ts_split]
    #     is_name, is_split = zip(*visnlangdatasetadapter.data_info[ts_split].items())
    #     visndatasetadapter = self.visndatasetadapterdict[is_name[0]][is_split[0][0]]
    #     img_info = visndatasetadapter.get(img_id)
    #     if vltk.features in img_info:
    #         feat = random.choice(img_info[vltk.features])
    #         return feat
    #     else:
    #         return None

    def transpose_vl(self, batch, max_size=512):
        visn_keys = tuple(self.visn)
        lang_keys = tuple(self.lang)
        if not visn_keys or not lang_keys:
            raise Exception(
                f"User must iterate through the  loader atleast once to use this function: {self.visn & self.lang},\
                {self.test}"
            )

        # first we resize image according to how many examples that we need
        val = None
        for x in lang_keys:
            val = batch.get(x, None)
            if val is not None:
                break
        n_exs_per_img = [len(i) for i in val]

        # process visn keys first
        device = None
        for visn_key in visn_keys:
            v = batch.get(visn_key, None)
            if v is None:
                continue
            if not isinstance(v, list):
                visn_val = torch.cat(
                    [
                        i.unsqueeze(0).expand(min(n, max_size), *i.shape)
                        for i, n in zip(v, n_exs_per_img)
                    ],
                    dim=0,
                )
                batch[visn_key] = visn_val
            else:
                visn_val = [[v] * min(n, max_size) for i, n in zip(v, n_exs_per_img)]
                batch[visn_key] = visn_val
            if device is not None:
                batch[visn_key] = batch[visn_key].to(device)

        # now we flatten the nested lang keys
        for lang_key in lang_keys:
            v = batch.get(lang_key, None)
            if v is None:
                continue
            if isinstance(v, list):
                if isinstance(v[0], torch.Tensor):
                    lang_val = torch.cat(
                        [j[: min(max_size, n)] for j, n in zip(v, n_exs_per_img)],
                        dim=0,
                    )
                    batch[lang_key] = lang_val
                elif isinstance(v[0], str):
                    lang_val = []
                    # here is also a part that we want to recude
                    for i, n in zip(v, n_exs_per_img):
                        if n >= max_size:
                            n = min(n, max_size)
                        lang_val.extend(i * n)
                    batch[lang_key] = lang_val
            else:
                lang_val = torch.stack(
                    [j[: min(max_size, n)] for j, n in zip(v, n_exs_per_img)],
                    dim=0,
                )
                batch[lang_key] = lang_val

        # raise Exception(batch)
        return batch

    def __len__(self):
        if self.config.img_first:
            return int(math.floor(len(self.uniq_imgs) * self.config.percent))
        else:
            return int(math.floor(len(self.datasets) * self.config.percent))

    @torch.no_grad()
    def __getitem__(self, i):
        if self.config.img_first:
            text_info, img_id = self._do_map_img_first(i)
            # raise Exception(len(self), img_id, i, self.uniq_imgs)
            (
                visnset_name,
                visnset_split,
            ) = self.visn_idx_organizer[i]
            img_info_dict_or_adapter = self.visndatasetadapterdict[visnset_name][
                visnset_split
            ]
            anno_dict_or_str = img_info_dict_or_adapter.get(img_id)
            if isinstance(anno_dict_or_str, str):
                anno_dict = {
                    vltk.filepath: anno_dict_or_str,
                    vltk.imgid: img_id,
                }
                anno_dict = self._handle_image(anno_dict)
            else:
                anno_dict = anno_dict_or_str

            # update with ground truth annotations, allow predcitions to overwrite GT
            if self.annotations is not None:
                annotations = self.annotations.get(img_id)
                annotations = {
                    k: v for k, v in annotations.items() if k not in anno_dict
                }
                anno_dict.update(annotations)

            anno_dict = self._handle_annotations(
                anno_dict,
                replace_keys=self.replace_keys,
            )
            text_info, anno_dict = self.run_visnlang_processors(
                text_info, anno_dict, self.config.img_first
            )

            entry = {**text_info, **anno_dict}
            # self.batch_info.update_visn_lang_keys(text_info, anno_dict)
            self.update_visn_lang_keys(text_info, anno_dict)
            entry = self.try_tensorify(entry)
            self.batch_info.update_entry_keys(entry)

            return entry

        else:
            # lets allow this to be the default
            text_info, img_id = self._do_map_text_first(i)
            (
                lang_split,
                langset_name,
                visnset_name,
                vinset_splits,
            ) = self.vl_idx_organizer[i]
            # TODO: for now I just check through all splits to see which imgid is present.
            # however, for all known datasets, there should only be one split present
            img_info_dict_or_filepath = None
            for vsplit in vinset_splits:
                visnadapter = self.visndatasetadapterdict[visnset_name][vsplit]
                try:
                    img_info_dict_or_filepath = visnadapter.get(img_id, None)
                except KeyError:
                    img_info_dict_or_filepath = None
                    continue

                if img_info_dict_or_filepath is None:
                    pass
                elif not isinstance(img_info_dict_or_filepath, (dict, str)):
                    img_info_dict_or_filepath = img_info_dict_or_filepath[0]
                    if not isinstance(img_info_dict_or_filepath, (dict, str)):
                        raise Exception(img_info_dict_or_filepath)

            if isinstance(img_info_dict_or_filepath, str):
                anno_dict = {
                    vltk.filepath: img_info_dict_or_filepath,
                    vltk.imgid: img_id,
                }
                anno_dict = self._handle_image(anno_dict)
            else:
                anno_dict = img_info_dict_or_filepath

            if self.annotations is not None:
                annotations = self.annotations.get(img_id)
                annotations = {
                    k: v for k, v in annotations.items() if k not in anno_dict
                }
                anno_dict.update(annotations)

            self._handle_annotations(
                anno_dict,
                replace_keys=self.replace_keys,
            )

            text_info, anno_dict = self.run_visnlang_processors(
                text_info, anno_dict, self.config.img_first
            )

            entry = {**text_info, **anno_dict}
            # self.batch_info.update_visn_lang_keys(text_info, anno_dict)
            self.update_visn_lang_keys(text_info, anno_dict)
            entry = self.try_tensorify(entry)
            self.batch_info.update_entry_keys(entry)
            return entry



import random

import numpy as np
import torch
from vltk.inspection import import_funcs_from_file
import vltk.vars as vltk


class Data:
    def __init__(self):
        if "DATADICT" not in globals():
            global DATADICT
            DATADICT = import_funcs_from_file(vltk.DATAPATH, pkg="vltk.processing")

    def avail(self):
        return list(DATADICT.keys())

    def get(self, name):
        return DATADICT[name]

    def add(self, name, lab):
        DATADICT[name] = lab


def one_hot_label(cur_entry, **kwargs):
    config = kwargs.get("config").lang

    label = cur_entry.get(vltk.label, None)
    score = cur_entry.get(vltk.score, None)
    if label is None:
        label = config.ignore_id
    elif label == config.ignore_id:
        cur_entry.pop(vltk.score, None)
        return
    else:
        if isinstance(label, int):
            pass
        elif len(label) == 1:
            label = label[0]
        else:
            score_sum = sum(score)
            prob = [ss / score_sum for ss in score]
            choice = np.random.multinomial(1, prob).argmax()
            label = label[choice]
    cur_entry.pop(vltk.score, None)
    cur_entry[vltk.label] = label


def multi_hot_label():
    pass


def masked_feature_modeling(cur_entry, **kwargs):
    if "roi_features" not in cur_entry:
        return
    random_feat = kwargs.get("random_feat_func")
    config = kwargs.get("config")
    mask_rate = config.feature_mask_rate
    if config.img_first:
        mask_rate /= 4
    roi_features = cur_entry["roi_features"]
    feat_mask = [0.0] * len(roi_features)
    feat_dim = len(roi_features[0])
    for i in range(len(roi_features)):
        prob = random.random()
        if prob < mask_rate:
            prob /= mask_rate
            # 80% randomly change token to zero feat
            if prob < 0.8:
                roi_features[i] = torch.zeros(feat_dim)

            # 10% randomly change token to random feat
            elif prob < 0.9:
                roi_features[i] = torch.tensor(random_feat())
            # Need to predict this feat
            feat_mask[i] = 1.0
    cur_entry["roi_features"] = roi_features
    cur_entry["feat_mask"] = torch.tensor(feat_mask)


def matched_sentence_modeling(cur_entry, **kwargs):
    random_sents = kwargs.get("random_sents")
    random_sent = lambda: random.choice(random_sents)
    config = kwargs.get("config")
    is_matched = 1
    text = cur_entry[vltk.text]
    rand_text = text
    if random.random() < config.sentence_match_rate:
        if vltk.label in cur_entry:
            cur_entry[vltk.label] = config.ignore_id
            cur_entry[vltk.score] = 0
        is_matched = 0
        while rand_text == text:
            rand_text = random_sent()

    cur_entry["is_matched"] = is_matched
    cur_entry[vltk.text] = rand_text
    return cur_entry


def masked_language_modeling(cur_entry, **kwargs):
    tokenizer = kwargs.get("tokenizer")
    config = kwargs.get("config")
    special_ids = kwargs.get("special_ids")
    n_ids = kwargs.get("n_ids")
    all_ids = kwargs.get("all_ids")
    random_id = lambda: all_ids[random.randint(0, n_ids - 1)]
    # special_ids = set([tokenizer.token_to_id(t) for t in kwargs.get("special_tokens")])
    # random_id = lambda: random.choice(list(tokenizer.get_vocab().items()))[1]

    input_ids = cur_entry["input_ids"]
    attention_mask = cur_entry["text_attention_mask"]
    ignore_id = config.ignore_id
    mask_id = tokenizer.token_to_id("[mask]")
    sep_id = tokenizer.token_to_id("[sep]")
    masked_labels = [ignore_id] * len(input_ids)
    masked_ids = input_ids
    for j, (iid, mid) in enumerate(zip(input_ids[1:], attention_mask[1:]), start=1):
        if int(mid) == 0 or iid == sep_id:
            break
        tid = random_id()
        while tid in special_ids:
            tid = random_id()
        mask_rate = config.word_mask_rate
        prob = random.random()
        if prob < mask_rate:
            old_id = masked_ids[j]
            prob /= mask_rate
            if prob < 0.8:
                masked_ids[j] = mask_id
                pass
            elif prob < 0.9:
                masked_ids[j] = tid
                pass
            masked_labels[j] = old_id
    cur_entry["input_ids"] = masked_ids
    cur_entry["masked_labels"] = masked_labels



import os
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from typing import Union

import torch
from vltk import LOOPPATH, 
# from vltk.dataset import UniversalLoader
from vltk.inspect import get_classes


class LoopIdentifier:
    pass


class Loop(LoopIdentifier, ABC):
    def __init__(
        self,
        config: object,
        datasets: str,
        label_dict: dict,
        visndatasetadapterdict: dict,
        visnlangdatasetadapterdict: dict,
        model_dict: Union[dict, None],
        extra_modules: Union[dict, None] = None,
    ):
        self.label_dict = label_dict
        self.model_dict = model_dict
        self.extra_modules = extra_modules
        self.config = config
        self.datasets = datasets
        self.cur_step = 0
        self.scheduler = None
        self.half_precision = getattr(config.train, "half_precision", False)
        self.main_device = (
            f"cuda:{config.gpu}" if getattr(config, "gpu", -1) != -1 else "cpu"
        )
        self.aux_model_device = (
            f"cuda:{config.aux_gpu}" if getattr(config, "aux_gpu", -1) != -1 else "cpu"
        )
        self.scaler = None if not self.half_precision else torch.cuda.amp.GradScaler()

        self._init_loader(visnlangdatasetadapterdict, visndatasetadapterdict, label_dict)
        self._init_models_and_extras(model_dict, extra_modules)
        self._init_optim()

        self._dataset = self.loader.dataset

        self.logdir = getattr(self.config, "logdir", None)
        if self.logdir is not None and self.config.logging:
            os.makedirs(self.logdir, exist_ok=True)

    def __key(self):
        name = self.is_name + "_"
        train = str(id(self.is_train)) + "_"
        if self.model_dict is None:
            models = ""
        else:
            models = "_".join(list(self.model_dict.keys())) + "_"
        if self.extra_modules is None:
            extras = ""
        else:
            extras = "_".join(list(self.extra_modules.keys())) + "_"
        return name + train + models + extras

    def __str__(self):
        return self.key()

    def __hash__(self):
        return hash(self.__key())

    # make MUCH better in the future
    def __eq__(self, other):
        if isinstance(other, Loop):
            if self.name == other.name and (
                (self.is_train and other.is_train)
                or (not self.is_train and not other.is_train)
            ):
                if self.extra_modules is None and other.extra_modules is None:
                    pass
                elif self.extra_modules is not None and other.extra_modules is None:
                    return False
                elif self.extra_modules is None and other.extra_modules is not None:
                    return False
                elif self.extra_modules.keys() != other.extra_modules.keys():
                    return False

                if self.model_dict is None:
                    if other.model_dict is None:
                        return True
                    else:
                        return False
                else:
                    if tuple(self.model_dict.keys()) == tuple(other.model_dict.keys()):
                        return True
                    else:
                        return False
                return True
            else:
                return False
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def _init_models_and_extras(self, model_dict, extra_modules=None):
        if model_dict is not None:
            for k, v in model_dict.items():
                v_old = getattr(self, k, None)
                assert v_old is None, (type(v_old), k)
                setattr(self, k, v)
        if extra_modules is not None:
            for k, v in extra_modules.items():
                v_old = getattr(self, k, None)
                assert v_old is None, type(v_old)
                v = v.to(self.main_device)
                setattr(self, k, v)

    @property
    def total_steps(self):
        if self.is_train:
            return self.config.train.epochs * len(self.loader)
        else:
            return len(self.loader)

    @property
    def warmup(self):
        return getattr(self, "_warmup", None)

    @property
    def optim(self):
        return getattr(self, "_optim", None)

    @property
    def dataset(self):
        return self._dataset

    def set_batch_size(self, batch):
        for k, v in batch.items():
            if isinstance(v, torch.Tensor):
                self._bz = v.size(0)
                break

    def get_grad_params(self):
        parameters = []
        for k, v in self.model_dict.items():
            parameters.extend([p for p in v.parameters() if p.requires_grad])
        if self.extra_modules is not None:
            for k, v in self.extra_modules.items():
                parameters.extend([p for p in v.parameters() if p.requires_grad])
        return parameters

    def tqdm_update(self, info: dict = None):
        if info is not None:
            clean_info = {}
            for k, v in info.items():
                if v is None or (not isinstance(v, bool) and not v):
                    continue
                else:
                    clean_info[k] = v

            self._tqdm.set_postfix(**clean_info)

    def write_iter(self, info: dict = None):
        if info is not None:
            clean_info = {}
            for k, v in info.items():
                if v is None or (not isinstance(v, bool) and not v):
                    continue
                else:
                    clean_info[k] = v
        logstr = ""
        for k, v in info.items():
            logstr += f"{k}={v}; "

        if self.config.logging and info is not None and info:
            logfile = os.path.join(self.config.logdir, "cur_epoch.txt")
            assert logfile is not None
            with open(logfile, "a") as f:
                date = datetime.datetime.now()
                f.write(f"Time: {date} | {logstr} \n")
                f.flush()
            return True
        return False

    def toCuda(self, batch, device=None):
        self.loader.toCuda(batch, device)

    def toTrain(self):
        for k, v in self.__dict__.items():
            if isinstance(v, torch.nn.Module):
                v.train()

    def toEval(self):
        for k, v in self.__dict__.items():
            if isinstance(v, torch.nn.Module):
                v.eval()

    # move to support multiple optimizers in the future for even more loop generality
    def step(self, loss=None):
        if self.is_train and loss is not None:
            if self.scaler is not None:
                self.scaler.scale(loss).backward()
                self.scaler.unscale_(self.optim)
                torch.nn..clip_grad_norm_(
                    self.get_grad_params(), self.config.train.max_norm
                )
                self.scaler.step(self.optim)
                self.scaler.update()
            else:
                loss.backward()
                torch.nn..clip_grad_norm_(
                    self.get_grad_params(), self.config.train.max_norm
                )
                self.optim.step()

            self.warmup.step()

    def get_lr(self):
        if not self.is_train:
            return []
        elif self.warmup is not None:
            return self.warmup.get_lr()
        else:
            lrs = []
            for param_group in self.optim.param_groups:
                lrs.append(param_group["lr"])
            return lrs

    def __iter__(self):
        for x in self.loader:
            yield x

    def __call__(self) -> dict:
        loop_outputs = defaultdict(list)
        if self.is_train:
            self.toTrain()
        else:
            self.toEval()
        with torch.no_grad() if not self.is_train else .dummy_context():
            for batch in self.tqdm:
                if self.is_train:
                    self.optim.zero_grad()
                self.set_batch_size(batch)
                with self.forward_context():
                    model_outputs = self.forward(batch)
                outputs = self.loop(batch, model_outputs)
                self.write_iter(outputs)

                if self.is_train:
                    losses = outputs.get("losses", None)
                    assert losses is not None
                    self.step(losses)
                self.cur_step += 1
                if outputs is not None and loop_outputs is not None:
                    for k, v in outputs.items():
                        loop_outputs[k].append(v)
                else:
                    loop_outputs = None
            return loop_outputs

    @property
    def split(self):
        return self._split

    def get_visnlangdatasetadapter(self):
        return self.loader.dataset.visnlangdatasetadapterdict

    def get_visndatasetadapter(self):
        return self.loader.dataset.visndatasetadapterdict

    def _init_loader(self, visnlangdatasetadapterdict, visndatasetadapterdict, label_dict):
        datasets = self.datasets if self.is_train else self.config.data.eval_datasets
        self.loader = UniversalLoader(
            config=self.config.data,
            names=datasets,
            label_dict=label_dict,
            visndatasetadapterdict=visndatasetadapterdict,
            visnlangdatasetadapterdict=visnlangdatasetadapterdict,
        )

    @classmethod
    def eval_instance(cls, eval_name):
        eval_cls = type(eval_name, (cls,), {"is_train": False, "name": eval_name})
        assert not eval_cls.is_train, "loop is already in eval mode"
        return eval_cls

    @abstractmethod
    def loop(self, batch, model_outputs) -> dict:
        return None

    @abstractmethod
    def forward(self, batch) -> object:
        return None

    @property
    @abstractmethod
    def name(self) -> str:
        return ""

    @property
    @abstractmethod
    def is_train(self):
        pass



import json
import os
from collections import defaultdict

import numpy as np
import torch
import vltk.vars as vltk
from matplotlib.patches import Rectangle
from PIL import Image
from pycocotools import mask as coco_mask
from torchvision.transforms.functional import resize
from tqdm import tqdm
from vltk.processing.image import get_pad, get_rawsize, get_scale, get_size

PATH = os.path.join(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..")), "libdata"
)
ANS_CONVERT = json.load(open(os.path.join(PATH, "convert_answers.json")))
CONTRACTION_CONVERT = json.load(open(os.path.join(PATH, "convert_answers.json")))


def map_ocr_predictions(pred, tokenmap, gold=None, boxes=None, ignore_id=-100):
    golds = []
    preds = []
    accs = []
    if gold is not None:
        for g, t, p in zip(gold, tokenmap, pred):
            t = t[: t.argmin()]
            total = 0
            for i, v in enumerate(t):
                total += v
                if total >= len(g):
                    break
            t = t[:i]

            tsum = sum(t)
            split_g = torch.split(g[:tsum], t.cpu().tolist())
            split_p = torch.split(p[:tsum], t.cpu().tolist())
            true_gold = torch.stack([x[0] for x in split_g]).cpu().tolist()
            true_preds = torch.stack([x.mode().values for x in split_p]).cpu().tolist()
            for i in reversed(range(len(true_gold))):
                if true_gold[i] == -100:
                    true_gold.pop(i)
                    true_preds.pop(i)
            accs += [(1 if p == g else 0) for p, g in zip(true_preds, true_gold)]
            preds += true_preds
            golds += true_gold
        return golds, preds, accs
    else:
        for t, p in zip(tokenmap, pred):
            t = t[: t.argmin()]
            total = 0
            for i, v in enumerate(t):
                total += v
                if total >= len(p):
                    break
            t = t[:i]

            tsum = sum(t)
            split_p = torch.split(p[:tsum], t.cpu().tolist())
            true_preds = torch.stack([x.mode().values for x in split_p]).cpu().tolist()
            preds += true_preds
        bboxes = None
        if boxes is not None:
            bboxes = []
            for t, b in zip(tokenmap, boxes):
                t = t[: t.argmin()]
                total = 0
                for i, v in enumerate(t):
                    total += v
                    if total >= len(b):
                        break
                t = t[:i]

                tsum = sum(t)
                split_b = torch.split(b[:tsum], t.cpu().tolist())
                temp_boxes = []
                for s in split_b:
                    temp_boxes.append(s.tolist()[0])
                true_boxes = temp_boxes
                bboxes += true_boxes
        return preds, bboxes


def histogram_from_counter(counter, truncate_labs=False, min_freq=0, x_label=""):
    # credits to stack overflow
    # TODO: add credits to stack overflow for other functions that I am using form there
    import matplotlib.pyplot as plt

    alt_msg = ""
    if "None" in counter:
        n_none = counter["None"]
        alt_msg = f"None: {n_none}"
        counter.pop("None")
    counter = sorted(list(counter.items()), key=lambda x: x[1])
    n_bins = len(counter)
    for i, (k, v) in enumerate(counter):
        if v >= min_freq:
            break
    counter = counter[i:]

    labels, values = zip(*counter)
    if truncate_labs:
        labels = [labels[0]] + ["" for x in labels[:-2]] + [labels[-1]]
    indexes = np.arange(len(labels))
    width = 1
    plt.bar(indexes, values, width)
    plt.xticks(indexes + width * 0.5, labels, rotation=-10)
    plt.ylabel("counts")
    plt.xlabel(x_label)
    plt.tight_layout()
    extra = Rectangle((0, 0), 1, 1, fc="w", fill=False, edgecolor="none", linewidth=0)
    plt.legend(
        [
            extra,
        ],
        (f"n_bins: {n_bins}; min_freq: {min_freq}" + f"; {alt_msg}",),
    )
    plt.show()


# def imagepoints_to_polygon(points):
#     img = imagepoints_to_mask(points)
#     polygon = mask_to_polygon(img)
#     return polygon


# source: https://github.com/ksrath0re/clevr-refplus-rec/
def imagepoints_to_mask(points, size):
    # raise Exception(points)
    # npimg = []
    img = []
    cur = 0
    for num in points:
        # if cur == 0:
        #     part = np.zeros(int(num))
        # else:
        #     part = np.concatenate((part, np.ones(int(num))))
        # npimg.append(part)
        num = int(num)
        img += [cur] * num
        cur = 1 - cur
    img = torch.tensor(img).reshape(tuple(size.tolist()))
    # npimg = np.stack(npimg)
    # raise Exception(part.shape)
    # part = part.reshape(tuple(size.tolist()))
    return img


# def mask_to_polygon(mask):
#     contours = measure.find_contours(mask, 0.5)
#     seg = []
#     for contour in contours:
#         contour = np.flip(contour, axis=1)
#         segmentation = contour.ravel().tolist()
#         seg.append(segmentation)
#     return seg


def rescale_box(boxes, wh_scale):
    # boxes = (n, (x, y, w, h))
    # x = top left x position
    # y = top left y position
    h_scale = wh_scale[1]
    w_scale = wh_scale[0]
    boxes[:, 0] *= w_scale
    boxes[:, 1] *= h_scale
    boxes[:, 2] *= w_scale
    boxes[:, 3] *= h_scale

    return boxes


def seg_to_mask(segmentation, w, h):
    segmentation = coco_mask.decode(coco_mask.frPyObjects(segmentation, h, w))
    if len(segmentation.shape) < 3:
        segmentation = segmentation[..., None]
    segmentation = np.any(segmentation, axis=-1).astype(np.uint8)
    return torch.from_numpy(segmentation).bool()


# def resize_mask(mask, transforms_dict):
#     if "Resize" in transforms_dict:
#         return transforms_dict["Resize"](mask)
#     else:
#         return mask


def resize_binary_mask(array, img_size, pad_size=None):
    img_size = (img_size[0], img_size[1])
    # raise Exception(img_size)
    if array.shape != img_size:
        if array.dim() == 2:
            array = array.unsqueeze(0).unsqueeze(0)
        else:
            array = array.unsqueeze(0)

        array = resize(array, img_size)

        if array.dim() == 2:
            array = array.squeeze(0).squeeze(0)
        else:
            array = array.squeeze(0)

        return array
    else:
        return torch.from_numpy(array)


def uncompress_mask(compressed, size):
    mask = np.zeros(size, dtype=np.uint8)
    mask[compressed[0], compressed[1]] = 1
    return mask


def clean_label(ans):
    if len(ans) == 0:
        return ""
    ans = ans.lower()
    ans = ans.replace(",", "")
    if ans[-1] == ".":
        ans = ans[:-1].strip()
    if ans.startswith("a "):
        ans = ans[2:].strip()
    if ans.startswith("an "):
        ans = ans[3:].strip()
    if ans.startswith("the "):
        ans = ans[4:].strip()
    ans = " ".join(
        [
            CONTRACTION_CONVERT[a] if a in CONTRACTION_CONVERT else a
            for a in ans.split(" ")
        ]
    )
    if ans in ANS_CONVERT:
        ans = ANS_CONVERT[ans]
    return ans


def soft_score(occurences):
    if occurences == 0:
        return 0
    elif occurences == 1:
        return 0.3
    elif occurences == 2:
        return 0.6
    elif occurences == 3:
        return 0.9
    else:
        return 1


def get_span_via_jaccard(words, answers, skipped=None):
    """
    inputs:
        words: tuple of strings (each string == one word)
        answers: list of strings (each string == one word or many words)
        skipped: int or None
    outputs:
        span: list of length `len(words)` of 1's and 0's. 0: is not apart of answer, 1: is apart of answer
        max_jaccard: the similarity metric value 0.0-1.0 for how well answer matched in span
    """
    span = [0] * len(words)
    any_ans = False
    for ans in answers:
        # single word case
        if len(ans.split()) == 1:
            try:
                idx = words.index(ans.lower())
            except Exception:
                continue
            if idx is not None:
                span[idx] = 1
                max_jaccard = 1.0
                any_ans = True
                break

    if not any_ans:
        keep = None
        max_jaccard = -0.1
        for ans in answers:
            if len(ans.split()) == 1:
                for ans in answers:
                    ans = set(ans.lower())
                    for idx, word in enumerate(words):
                        word = set(word.lower())
                        jaccard = len(word & ans) / len(word | ans)
                        if jaccard > max_jaccard:
                            max_jaccard = jaccard
                            keep = idx
            else:
                end_keep = len(words)

                ans = ans.split()
                start_keep = end_keep - len(ans)
                start = ans[0].lower()
                end = ans[-1].lower()
                start = set(start)
                end = set(end)
                jaccards = []
                for idx, word in enumerate(words[: -len(ans)]):
                    temp_jaccard = 0.0
                    for jdx, subans in enumerate(ans):
                        word = set(words[idx + jdx].lower())
                        subans = set(subans)
                        temp_jaccard += len(word & subans) / len(word | subans)
                        jaccards.append(
                            (
                                temp_jaccard / len(ans),
                                (idx, idx + len(ans)),
                            )
                        )
                if not jaccards:
                    continue
                jaccard, (start_keep, end_keep) = sorted(jaccards, key=lambda x: x[0])[
                    -1
                ]
                if jaccard > max_jaccard:
                    keep = (start_keep, end_keep)
                    span[start_keep:end_keep] = [1] * ((end_keep - start_keep) + 1)
                    max_jaccard = jaccard

        if keep is None:
            span = None
        elif isinstance(keep, tuple):
            span[keep[0] : keep[1]] = [1] * ((keep[1] - keep[0]) + 1)
        else:
            span[keep] = 1

    if max_jaccard == 0.0:
        if skipped is not None:
            skipped += 1
        span = None

    return span, max_jaccard


def truncate_and_pad_list(inp_list, max_len, pad_value=""):
    inp_list = inp_list[: min(max_len, len(inp_list))]
    inp_list += [pad_value] * (max_len - len(inp_list))
    return inp_list


def basic_coco_annotations(json_files, splits):
    """
    inputs:
        json_files: a dict of annotation files in  coco format -->
            keys: filename
            values: json file
            ===
            boxes in (x,y,w,h) format
            segmentations are polygons
            ...
        splits: list of respective splits aligned to the keys of json_files
    outputs:
        list of dictionaries:
            keys:
                vltk.imgid
                vltk.box
                vltk.polygons
                vltk.objects
            values:
                str: the image id (stem of the filename)
                list of list of floats: list of bounding boxes
                list of list of list of floats: list of polygons
                list of strings: respective label classes for each object/segmentation
    """
    total_annos = {}
    id_to_cat = {}
    file_to_id_to_stem = defaultdict(dict)
    for file, data in sorted(json_files.items(), key=lambda x: x[0]):
        info = data["images"]
        for i in info:
            img_id = i["file_name"].split(".")[0]

            file_to_id_to_stem[file][i["id"]] = img_id

        # for file, data in tqdm(sorted(json_files.items(), key=lambda x: x[0])):
        categories = data["categories"]
        for cat in categories:
            id_to_cat[cat["id"]] = cat["name"]

        for entry in data["annotations"]:

            img_id = str(file_to_id_to_stem[file][entry["image_id"]])
            bbox = entry["bbox"]
            segmentation = entry["segmentation"]
            category_id = id_to_cat[entry["category_id"]]
            if entry["iscrowd"]:
                seg_mask = []
            else:
                seg_mask = segmentation
                if not isinstance(seg_mask[0], list):
                    seg_mask = [seg_mask]
            img_data = total_annos.get(img_id, None)
            if img_data is None:
                img_entry = defaultdict(list)
                img_entry[vltk.objects].append(category_id)
                img_entry[vltk.box].append(bbox)
                img_entry[vltk.polygons].append(seg_mask)
                total_annos[img_id] = img_entry
            else:
                total_annos[img_id][vltk.box].append(bbox)
                total_annos[img_id][vltk.objects].append(category_id)
                total_annos[img_id][vltk.polygons].append(seg_mask)

    return [{vltk.imgid: img_id, **entry} for img_id, entry in total_annos.items()]



from vltk import SCHEDPATH
from vltk.inspection import import_funcs_from_file


class Sched:
    def __init__(self):
        if "SCHEDDICT" not in globals():
            global SCHEDDICT
            SCHEDDICT = import_funcs_from_file(SCHEDPATH, pkg="vltk.processing")

    def avail(self):
        return list(SCHEDDICT.keys())

    def get(self, name):
        return SCHEDDICT[name]

    def add(self, name, lab):
        SCHEDDICT[name] = lab
